import tensorflow as tf
from tensorflow.python.ops import array_ops

EPSILON = 1e-8

def sigmoid_focal_loss (labels, logits, alpha = 0.25, gamma = 2):    
    p = tf.nn.sigmoid (logits)
    zeros = array_ops.zeros_like (p, dtype = p.dtype)
    # For poitive prediction, only need consider front part loss, back part is 0;
    # target_tensor > zeros <=> z=1, so poitive coefficient = z - p.
    pos_p_sub = array_ops.where (labels > zeros, labels - p, zeros)    
    # For negative prediction, only need consider back part loss, front part is 0;
    # target_tensor > zeros <=> z=1, so negative coefficient = 0.
    neg_p_sub = array_ops.where (labels > zeros, zeros, p)
    per_entry_cross_ent = - alpha * (pos_p_sub ** gamma) * tf.log (tf.clip_by_value (p, EPSILON, 1.0)) \
                          - (1 - alpha) * (neg_p_sub ** gamma) * tf.log (tf.clip_by_value(1.0 - p, EPSILON, 1.0))
    return tf.reduce_mean (tf.reduce_sum (per_entry_cross_ent, axis = 1))

def categorical_focal_loss (labels, logits, alpha = 0.25, gamma = 2):  
  # Scale predictions so that the class probas of each sample sum to 1
  logits = tf.nn.softmax (logits)
  logits /= tf.reduce_sum (logits, axis=-1, keepdims=True)
  # Clip the prediction value to prevent NaN's and Inf's
  logits = tf.clip_by_value (logits, EPSILON, 1. - EPSILON)
  # Calculate Cross Entropy
  cross_entropy = -labels * tf.log (logits)
  # Calculate Focal Loss
  loss = alpha * tf.pow (1 - logits, gamma) * cross_entropy

  # Sum the losses in mini_batch
  return tf.reduce_mean (tf.reduce_sum (loss, axis = 1))

def binary_focal_loss (labels, logits, alpha = 0.25, gamma = 2): 
  logits = tf.nn.sigmoid (logits)
  pt_1 = tf.where(tf.equal(labels, 1), logits, tf.ones_like(logits))
  pt_0 = tf.where(tf.equal(labels, 0), logits, tf.zeros_like(logits))

  # clip to prevent NaN's and Inf's
  pt_1 = tf.clip_by_value (pt_1, EPSILON, 1. - EPSILON)
  pt_0 = tf.clip_by_value (pt_0, EPSILON, 1. - EPSILON)

  return tf.reduce_mean (
          -tf.reduce_sum (alpha * tf.pow (1. - pt_1, gamma) * tf.log (pt_1))
          -tf.reduce_sum ((1 - alpha) * tf.pow (pt_0, gamma) * tf.log (1. - pt_0))
  )



import librosa
import numpy as np
import random

def _gen_colored_noise (spectral_shape, samples_to_generate):
    # https://www.kaggle.com/holzner/generating-different-colors-of-noise
    def normalize(samples):
        std = samples.std()
        if std > 0:
            return samples / std
        else:
            return samples
    
    flat_spectrum = np.random.normal(size = samples_to_generate // 2 + 1) + \
            1j * np.random.normal(size = samples_to_generate // 2 + 1)
    return normalize(np.fft.irfft (flat_spectrum * spectral_shape).real)

COLORS = [
    'white',   # flat spectrum
    'pink',    # -3dB (factor 0.5)  per octave / -10 dB (factor 0.1) per decade
    'blue',    # +3dB (factor 2)    per octave / +10 dB (factor 10) per decade
    'brown',   # -6dB (factor 0.25) per octave / -20 dB (factor 0.01) per decade
    'violet'
]
NOISE_CACHE = {}
def _gen_noise (color, samples_to_generate):
    if color == 'white':
        return np.random.normal(size = samples_to_generate)    
    spectrum_len = samples_to_generate // 2 + 1
    if color == 'pink':
        return _gen_colored_noise(1. / (np.sqrt(np.arange(spectrum_len) + 1.)), samples_to_generate)        
    elif color == 'blue':
        return _gen_colored_noise(np.sqrt(np.arange(spectrum_len)), samples_to_generate)    
    elif color == 'brown' or color == 'red':
        return _gen_colored_noise(1. / (np.arange(spectrum_len) + 1), samples_to_generate)
    elif color == 'violet' or color == 'purple':
        return _gen_colored_noise(np.arange(spectrum_len), samples_to_generate)    

for color in COLORS:    
    NOISE_CACHE [color] = _gen_noise (color, 16000 * 180)

def gen_noise (color, samples_to_generate):
    global NOISE_CACHE

    assert samples_to_generate % 2 == 0
    if samples_to_generate == len (NOISE_CACHE [color]):
        return NOISE_CACHE [color]
    if samples_to_generate <= len (NOISE_CACHE [color]):
        start = random.randrange (0, len (NOISE_CACHE [color]) - samples_to_generate)
        return NOISE_CACHE [color][start:start + samples_to_generate]
    return  _gen_noise (color, samples_to_generate)

def add_noise (y, sr, min_gain = 0.01, max_gain = 0.12):
    if y.shape [0] % 2 == 1:
        y = y [:-1] # make even

    noise = gen_noise (random.choice (COLORS), y.shape [0]) / random.randrange (10, 50)
    if random.randrange (3) == 0:
        noise += gen_noise (random.choice (COLORS), y.shape [0]) / random.randrange (10, 50)
        noise /= 2

    gain = np.random.uniform (min_gain, max_gain)    
    return y + ((gain * noise).astype ("float32"))
    
def speed (y, sr, max_rate = 0.1):
    rate = 1.0
    if random.randrange (2):
        rate += max (0.02, random.random () / (max_rate * 100))
    else:
        rate -= max (0.02, random.random () / (max_rate * 100))     
    return librosa.effects.time_stretch (y, rate)    
    
def pitch_shift (y, sr, max_step = 0.5):
    return librosa.effects.pitch_shift (y, sr, random.random () * max_step * random.choice ([1, -1]), bins_per_octave = 12)        

def random_laps (y, sr):
    if librosa.get_duration(y, sr = sr) > 1.6:
       time_lap = random.random () / 3
       removable = int (sr * time_lap)
       y = y [removable:len (y) - removable]
       y = np.roll (y, random.randrange (len (y)))    
    return y   



# Noise data from:
#    https://github.com/MycroftAI/Precise-Community-Data/blob/master/not-wake-words/noises/noises-20190814-elt.tar.gz

from . import feature
from glob import glob
import numpy as np
import os
from random import shuffle

class NoiseData:
    def __init__(self, noise_folder, sr = 16000):
        self.noise_data = [
            feature.load (file, sr) [0]
            for file in glob (os.path.join(noise_folder, '*.wav'))
        ]
        self.noise_data_id = 0
        self.noise_pos = 0
        self.repeat_count = 0

    def get_fresh_noise(self, n) -> np.ndarray:
        noise_audio = np.empty(0)
        while len(noise_audio) < n:
            noise_source = self.noise_data[self.noise_data_id]
            noise_chunk = noise_source[self.noise_pos:self.noise_pos + n - len(noise_audio)]
            self.noise_pos += n - len(noise_audio)
            if self.noise_pos >= len(noise_source):
                self.noise_pos = 0
                self.noise_data_id += 1
                if self.noise_data_id >= len(self.noise_data):
                    self.noise_data_id = 0
                    shuffle (self.noise_data)
                    self.repeat_count += 1
                    if self.repeat_count == 100:
                        print ('Warning: Repeating noise 100+ times. Add more to prevent '
                              'overfitting.')
            noise_audio = np.concatenate([noise_audio, noise_chunk])
        return noise_audio

    def compose (self, audio, noise_ratio) -> np.ndarray:
        noise_data = self.get_fresh_noise(len(audio))
        audio_volume = feature.get_volume (audio)
        noise_volume = feature.get_volume (noise_data)        
        adjusted_noise = audio_volume * noise_data / noise_volume
        return noise_ratio * adjusted_noise + (1.0 - noise_ratio) * audio

def create (noise_folder, sr = 16000):
    return NoiseData (noise_folder, sr)



import wave
import time
import logging

DEFAULT_AUDIO_SAMPLE_RATE = 16000
DEFAULT_AUDIO_SAMPLE_WIDTH = 2
DEFAULT_AUDIO_DEVICE_BLOCK_SIZE = 6400
DEFAULT_AUDIO_DEVICE_FLUSH_SIZE = 25600

class WaveSource:
    def __init__(self, fp, sample_rate = DEFAULT_AUDIO_SAMPLE_RATE, sample_width = DEFAULT_AUDIO_SAMPLE_WIDTH, mimic = True, pad_silent = 0):
        self._fp = fp
        try:
            self._wavep = wave.open(self._fp, 'r')
        except wave.Error as e:
            logging.warning('error opening WAV file: %s, '
                            'falling back to RAW format', e)
            self._fp.seek(0)
            self._wavep = None
        self._sample_rate = sample_rate
        self._sample_width = sample_width
        self._sleep_until = 0
        self._mimic = mimic
        self._silence = pad_silent

    def read(self, size):
        now = time.time()
        missing_dt = (self._sleep_until - now)
        if self._mimic and missing_dt > 0:
            time.sleep(missing_dt)
        self._sleep_until = time.time() + self._sleep_time(size)
        data = (self._wavep.readframes(size // self._sample_width)
                if self._wavep
                else self._fp.read(size))

        if not data:
            self._silence -= size
            if self._silence <= 0:
                return b''
            return b'\x00' * size

        if len (data) < size:
            return data + (b'\x00' * (size - len (data)))
        return data

    def write(self, data):
        self._wavep.writeframes(data)

    def close(self):
        if self._wavep:
            self._wavep.close()
        self._fp.close()

    def start (self):
        pass

    def stop(self):
        pass

    def _sleep_time(self, size):
        sample_count = size / float(self._sample_width)
        sample_rate_dt = sample_count / float(self._sample_rate)
        return sample_rate_dt

    def flush(self):
        pass

    @property
    def sample_rate(self):
        return self._sample_rate


try:
    import sounddevice as sd
except ImportError:
    pass

class SoundDeviceStream (WaveSource):
    def __init__(self, sample_rate = DEFAULT_AUDIO_SAMPLE_RATE, sample_width = DEFAULT_AUDIO_SAMPLE_WIDTH, block_size = DEFAULT_AUDIO_DEVICE_BLOCK_SIZE, flush_size = DEFAULT_AUDIO_DEVICE_FLUSH_SIZE, device = None):
        try:
            sd
        except NameError:
            raise Exception('sounddevice is not installed, use pip')

        if sample_width == 2:
            audio_format = 'int16'
        else:
            raise Exception('unsupported sample width:', sample_width)
        self._audio_stream = sd.RawStream(
            samplerate=sample_rate, dtype=audio_format, channels=1, device = device,
            blocksize=int(block_size/2),  # blocksize is in number of frames.
        )
        self._sample_width = sample_width
        self._block_size = block_size
        self._flush_size = flush_size
        self._sample_rate = sample_rate

    def read(self, size):
        size = size // self._sample_width # to bytes length
        buf, overflow = self._audio_stream.read(size)
        if overflow:
            logging.warning('SoundDeviceStream read overflow (%d, %d)', size, len(buf))
        return bytes(buf)

    def write(self, buf):
        underflow = self._audio_stream.write(buf)
        if underflow:
            logging.warning('SoundDeviceStream write underflow (size: %d)', len(buf))
        return len(buf)

    def flush(self):
        if self._audio_stream.active and self._flush_size > 0:
            self._audio_stream.write(b'\x00' * self._flush_size)

    def start(self):
        if not self._audio_stream.active:
            self._audio_stream.start()

    def stop(self):
        if self._audio_stream.active:
            self._audio_stream.stop()

    def close(self):
        if self._audio_stream:
            self.stop()
            self._audio_stream.close()
            self._audio_stream = None


import numpy as np
import sys
import os
import glob
import wave
import multiprocessing
from functools import partial
from scipy.fftpack import fft
from scipy.io import wavfile
from scipy import signal
from tqdm import tqdm
import librosa
import librosa.util
import random 
from scipy.stats import skew, kurtosis
import warnings
import resampy
from . import augments

SAMPLE_RATE = 16000

def _save (data, wavfile, target_path):
    savedir = os.path.join (target_path, os.path.basename (os.path.dirname (os.path.dirname (wavfile))))
    if not os.path.isdir (savedir):
        os.mkdir (savedir)
    target_file = os.path.join(savedir, "%s.npy" % os.path.basename (wavfile))
    np.save(target_file, data)

def get_volume (y):
    return np.sqrt (np.sum (y ** 2))

def norm_volmue (y):
    return y / np.max (np.abs (y))
_normalize = norm_volmue

def whitening (y):
    mean = np.mean (y)    
    std = np.std (y)  
    return (y - mean) / std
    
def log (x, name = ""):
    return np.log (np.abs (x) + 1e-3) / 6.7
  
def add_padding (x, padding):
    if not padding:
        return x
    return np.concatenate ([x, np.zeros (padding)])

def to_db (S):
  S = np.abs (S) + 1e-10 
  return 10 * np.log10(S) - 10 * np.log10(np.max (S))

def gain (input, output, target_dBFS = -20.0):
    from pydub import AudioSegment
    
    sound = AudioSegment.from_file (input)
    change_in_dBFS = target_dBFS - sound.dBFS
    normalized_sound = sound.apply_gain (change_in_dBFS)
    normalized_sound.export(output, output.split (".")[-1].lower ())
    return normalized_sound.dBFS
        
def _compress (data, name, padding = 0, valid_max = 2.0, valid_min = -2.0):
    # padding 3 means 2 convolution, 1 polling    
    data = data.real
    if name == "poly_features":
        data = data / 10.
    elif name == "rmse":
        data = data / 4.
    elif name == "spectral_contrast":
        data = data / 50.
    elif name == "mel":
        data = data / -80.
    elif name in ("stft", "spectral_centroid", "spectral_bandwidth", "spectral_rolloff"):        
        data = log (data, name)
                          
    features = np.array ([
        add_padding (np.mean (data, axis = 1), padding),
        add_padding (np.max (data, axis = 1), padding),
        add_padding (np.min (data, axis = 1), padding),
        add_padding (np.median (data, axis = 1), padding),
        add_padding (np.var (data, axis = 1), padding),
        add_padding (skew (data, axis = 1) / 15., padding),
        add_padding (np.log (abs (kurtosis (data, axis = 1)) + 1e-3) / 6.7, padding)
    ])
    if np.min (features) < valid_min or np.max (features) > valid_max:
        #print (name, int (np.max (features)), int (np.max (features)))
        raise AssertionError
    return features

def augment (y, sr, time_stretch = False, pitch_shift = False, random_laps = False, add_noise = False):    
    if add_noise:
        y = augments.add_noise (y, sr, 0.01, 0.12)            
    if time_stretch:
        y = augments.speed (y, sr, 0.1)
    if pitch_shift:
        y = augments.pitch_shift (y, sr, 0.5)    
    if random_laps and librosa.get_duration(y, sr = sr) > 1.6:
       y = augments.random_laps (y, sr)
    return y   

def augment_with_norm (y, sr = SAMPLE_RATE, *args, **karg):
    y = augment (y, sr, *args, **karg)
    y = norm_volmue (y)
    return whitening (y)

def loadraw (wavfile, sample_rate = SAMPLE_RATE, time_lap = 0.2, due_limit = (0.5, 10), trim_db = 60, res_type = "kaiser_best"):
    if res_type == "hann_window":
        y, sr_orig = librosa.load (wavfile, None, True)
        y = resampy.resample(y, sr_orig, sample_rate, filter = 'sinc_window', window = signal.hann)
        sr = sample_rate
    else:
        y, sr = librosa.load (wavfile, sample_rate, True, res_type = res_type)
    
    if time_lap:
        removable = int (sample_rate * time_lap)
        y = y [removable:len (y) - removable]
    # Trim the beginning and ending silence
    if trim_db:
        try:
            y, index = librosa.effects.trim (y, trim_db)
        except ValueError:        
            return None, sr   
    if due_limit:    
        duration = librosa.get_duration(y, sr = sr)
        #print ("    + duration: {:2.1f} sec => {:2.1f} sec".format (a_duration, duration))
        if duration < due_limit [0] or duration > due_limit [1]:
            warnings.warn ("audio file length is invalid, ignored")
            return None, sr
    if len (y) % 2 == 1:
        y = y [:-1]
    return y, sr

def load (wavfile, sample_rate = SAMPLE_RATE, res_type = "kaiser_best"):
    return loadraw (wavfile, sample_rate, None, None, 0, res_type)
    

if __name__ == '__main__':    
    pass



#-------------------------------------------------------------------------------
# Original source from: 
#   https://github.com/tensorlayer/tensorlayer/blob/master/tensorlayer/cost.py
#-------------------------------------------------------------------------------


import logging
import tensorflow as tf

_FLOATX = tf.float32
_EPSILON = 1e-8

def sparse_softmax_cross_entropy (output, target, name=None):    
    return tf.reduce_mean (tf.nn.sparse_softmax_cross_entropy_with_logits (labels=target, logits=output, name=name))

def categorical_crossentropy (logits, labels):    
    logits = tf.nn.softmax (logits)
    logits /= tf.reduce_sum(logits,
                            reduction_indices = len (logits.get_shape()) - 1,
                            keepdims=True)
    # manual computation of crossentropy
    logits = tf.clip_by_value(logits, tf.cast(_EPSILON, dtype=_FLOATX),
                              tf.cast(1.-_EPSILON, dtype=_FLOATX))
    cross_entropy = - tf.reduce_sum(labels * tf.log(logits),
                           reduction_indices=len(logits.get_shape())-1)
    return tf.reduce_mean(cross_entropy)


def sigmoid_cross_entropy (output, target, name=None):
    return tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(labels=target, logits=output, name=name))

def binary_cross_entropy (output, target, epsilon=1e-8, name='bce_loss'):
    with tf.name_scope(name):
        return tf.reduce_mean(tf.reduce_sum(-(target * tf.log(output + epsilon) + (1. - target) * tf.log(1. - output + epsilon)), axis=1))

def mean_squared_error(output, target, is_mean=False, name="mean_squared_error"):    
    with tf.name_scope(name):
        if output.get_shape().ndims == 2:  # [batch_size, n_feature]
            if is_mean:
                mse = tf.reduce_mean(tf.reduce_mean(tf.squared_difference(output, target), 1))
            else:
                mse = tf.reduce_mean(tf.reduce_sum(tf.squared_difference(output, target), 1))
        elif output.get_shape().ndims == 3:  # [batch_size, w, h]
            if is_mean:
                mse = tf.reduce_mean(tf.reduce_mean(tf.squared_difference(output, target), [1, 2]))
            else:
                mse = tf.reduce_mean(tf.reduce_sum(tf.squared_difference(output, target), [1, 2]))
        elif output.get_shape().ndims == 4:  # [batch_size, w, h, c]
            if is_mean:
                mse = tf.reduce_mean(tf.reduce_mean(tf.squared_difference(output, target), [1, 2, 3]))
            else:
                mse = tf.reduce_mean(tf.reduce_sum(tf.squared_difference(output, target), [1, 2, 3]))
        else:
            raise Exception("Unknow dimension")
        return mse

def normalized_mean_squared_error(output, target):
    with tf.name_scope("mean_squared_error_loss"):
        if output.get_shape().ndims == 2:  # [batch_size, n_feature]
            nmse_a = tf.sqrt(tf.reduce_sum(tf.squared_difference(output, target), axis=1))
            nmse_b = tf.sqrt(tf.reduce_sum(tf.square(target), axis=1))
        elif output.get_shape().ndims == 3:  # [batch_size, w, h]
            nmse_a = tf.sqrt(tf.reduce_sum(tf.squared_difference(output, target), axis=[1, 2]))
            nmse_b = tf.sqrt(tf.reduce_sum(tf.square(target), axis=[1, 2]))
        elif output.get_shape().ndims == 4:  # [batch_size, w, h, c]
            nmse_a = tf.sqrt(tf.reduce_sum(tf.squared_difference(output, target), axis=[1, 2, 3]))
            nmse_b = tf.sqrt(tf.reduce_sum(tf.square(target), axis=[1, 2, 3]))
        nmse = tf.reduce_mean(nmse_a / nmse_b)
    return nmse

def absolute_difference_error(output, target, is_mean=False):
    # is_mean : boolean Whether compute the mean or sum for each example.
    with tf.name_scope("mean_squared_error_loss"):
        if output.get_shape().ndims == 2:  # [batch_size, n_feature]
            if is_mean:
                loss = tf.reduce_mean(tf.reduce_mean(tf.abs(output - target), 1))
            else:
                loss = tf.reduce_mean(tf.reduce_sum(tf.abs(output - target), 1))
        elif output.get_shape().ndims == 3:  # [batch_size, w, h]
            if is_mean:
                loss = tf.reduce_mean(tf.reduce_mean(tf.abs(output - target), [1, 2]))
            else:
                loss = tf.reduce_mean(tf.reduce_sum(tf.abs(output - target), [1, 2]))
        elif output.get_shape().ndims == 4:  # [batch_size, w, h, c]
            if is_mean:
                loss = tf.reduce_mean(tf.reduce_mean(tf.abs(output - target), [1, 2, 3]))
            else:
                loss = tf.reduce_mean(tf.reduce_sum(tf.abs(output - target), [1, 2, 3]))
        else:
            raise Exception("Unknow dimension")
        return loss

def cosine_distance_error (output, target):
    return tf.reduce_mean (
        tf.losses.cosine_distance (
            tf.nn.l2_normalize(output, 1), 
            tf.nn.l2_normalize(target, 1),
            1 # dim / axis
        )
    )

def dice_coe (output, target, loss_type='jaccard', axis=(1, 2, 3), smooth=1e-5):
    inse = tf.reduce_sum(output * target, axis=axis)
    if loss_type == 'jaccard':
        l = tf.reduce_sum(output * output, axis=axis)
        r = tf.reduce_sum(target * target, axis=axis)
    elif loss_type == 'sorensen':
        l = tf.reduce_sum(output, axis=axis)
        r = tf.reduce_sum(target, axis=axis)
    else:
        raise Exception("Unknow loss_type")
    return tf.reduce_mean((2. * inse + smooth) / (l + r + smooth))

def dice_hard_coe (output, target, threshold=0.5, axis=(1, 2, 3), smooth=1e-5):
    output = tf.cast(output > threshold, dtype=tf.float32)
    target = tf.cast(target > threshold, dtype=tf.float32)
    inse = tf.reduce_sum(tf.multiply(output, target), axis=axis)
    l = tf.reduce_sum(output, axis=axis)
    r = tf.reduce_sum(target, axis=axis)
    return tf.reduce_mean((2. * inse + smooth) / (l + r + smooth))
    
def iou_coe (output, target, threshold=0.5, axis=(1, 2, 3), smooth=1e-5):
    pre = tf.cast(output > threshold, dtype=tf.float32)
    truth = tf.cast(target > threshold, dtype=tf.float32)
    inse = tf.reduce_sum(tf.multiply(pre, truth), axis=axis)  # AND
    union = tf.reduce_sum(tf.cast(tf.add(pre, truth) >= 1, dtype=tf.float32), axis=axis)  # OR
    return tf.reduce_mean((inse + smooth) / (union + smooth))
    
def cross_entropy_seq (logits, target_seqs, batch_size=None):  #, batch_size=1, num_steps=None):
    sequence_loss_by_example_fn = tf.contrib.legacy_seq2seq.sequence_loss_by_example
    loss = sequence_loss_by_example_fn([logits], [tf.reshape(target_seqs, [-1])], [tf.ones_like(tf.reshape(target_seqs, [-1]), dtype=tf.float32)])
    # [tf.ones([batch_size * num_steps])])
    cost = tf.reduce_sum(loss)  #/ batch_size
    if batch_size is not None:
        cost = cost / batch_size
    return cost

def cosine_similarity (v1, v2):
    return (tf.reduce_sum(tf.multiply(v1, v2), 1) 
        / (tf.sqrt(tf.reduce_sum(tf.multiply(v1, v1), 1)) 
        * tf.sqrt(tf.reduce_sum(tf.multiply(v2, v2), 1))))

def mean_seq (logits, target_seqs, batch_size, sequence_length):
    weights = tf.ones ([batch_size, sequence_length])
    sequence_loss = tf.contrib.seq2seq.sequence_loss (
        logits = logits, targets = target_seqs, weights = weights
    )
    return tf.reduce_mean (sequence_loss)

def cross_entropy_seq_with_mask (logits, target_seqs, input_mask, return_details=False, name=None):
    targets = tf.reshape(target_seqs, [-1])  # to one vector
    weights = tf.to_float(tf.reshape(input_mask, [-1]))  # to one vector like targets
    losses = tf.nn.sparse_softmax_cross_entropy_with_logits(logits=logits, labels=targets, name=name) * weights
    
    loss = tf.divide(
        tf.reduce_sum(losses),  # loss from mask. reduce_sum before element-wise mul with mask !!
        tf.reduce_sum(weights),
        name="seq_loss_with_mask")
    if return_details:
        return loss, losses, weights, targets
    else:
        return loss
